# Copyright 2019 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List
from typing import Set
from typing import Tuple

try:
    import importlib.metadata as importlib_metadata
except ModuleNotFoundError:
    import importlib_metadata

from ros2cli.node.strategy import NodeStrategy
from ros2doctor.api.format import doctor_warn


class DoctorCheck:
    """Abstract base class of ros2doctor check."""

    def category(self) -> str:
        """:return: string linking checks and reports."""
        raise NotImplementedError

    def check(self) -> bool:
        """:return: boolean indicating result of checks."""
        raise NotImplementedError


class DoctorReport:
    """Abstract base class of ros2doctor report."""

    def category(self) -> str:
        """:return: string linking checks and reports."""
        raise NotImplementedError

    def report(self) -> 'Report':  # using str as wrapper for custom class Report
        """:return: Report object storing report content."""
        raise NotImplementedError


class Report:
    """Stores report name and content."""

    __slots__ = ['name', 'items']

    def __init__(self, name: str):
        """Initialize with report name."""
        self.name = name
        self.items = []

    def add_to_report(self, item_name: str, item_info: str) -> None:
        """Add report content to items list (list of string tuples)."""
        self.items.append((item_name, item_info))


class Result:
    """Stores check result."""

    __slots__ = ['error', 'warning']

    def __init__(self):
        """Initialize with no error or warning."""
        self.error = 0
        self.warning = 0

    def add_error(self):
        self.error += 1

    def add_warning(self):
        self.warning += 1


def run_checks(*, include_warnings=False) -> Tuple[Set[str], int, int]:
    """
    Run all checks and return check results.

    :return: 3-tuple (categories of failed checks, number of failed checks,
             total number of checks)
    """
    fail_categories = set()  # remove repeating elements
    fail = 0
    total = 0
    entry_points = importlib_metadata.entry_points()
    if hasattr(entry_points, 'select'):
        groups = entry_points.select(group='ros2doctor.checks')
    else:
        groups = entry_points.get('ros2doctor.checks', [])
    for check_entry_pt in groups:
        try:
            check_class = check_entry_pt.load()
        except ImportError as e:
            doctor_warn(f'Check entry point {check_entry_pt.name} fails to load: {e}')
            continue
        try:
            check_instance = check_class()
        except Exception:
            doctor_warn(f'Unable to instantiate check object from {check_entry_pt.name}.')
            continue
        try:
            check_category = check_instance.category()
            result = check_instance.check()
            if result.error or (include_warnings and result.warning):
                fail += 1
                fail_categories.add(check_category)
            total += 1
        except Exception:
            doctor_warn(f'Fail to call {check_entry_pt.name} class functions.')
    return fail_categories, fail, total


def generate_reports(*, categories=None) -> List[Report]:
    """
    Print all reports or reports of failed checks to terminal.

    :return: list of Report objects
    """
    reports = []
    entry_points = importlib_metadata.entry_points()
    if hasattr(entry_points, 'select'):
        groups = entry_points.select(group='ros2doctor.report')
    else:
        groups = entry_points.get('ros2doctor.report', [])
    for report_entry_pt in groups:
        try:
            report_class = report_entry_pt.load()
        except ImportError as e:
            doctor_warn(f'Report entry point {report_entry_pt.name} fails to load: {e}')
            continue
        try:
            report_instance = report_class()
        except Exception:
            doctor_warn(f'Unable to instantiate report object from {report_entry_pt.name}.')
            continue
        try:
            report_category = report_instance.category()
            report = report_instance.report()
            if categories:
                if report_category in categories:
                    reports.append(report)
            else:
                reports.append(report)
        except Exception:
            doctor_warn(f'Fail to call {report_entry_pt.name} class functions.')
    return reports


def get_topic_names(skip_topics: List = ()) -> List:
    """Get all topic names using rclpy API."""
    topics = []
    with NodeStrategy(None) as node:
        topic_names_types = node.get_topic_names_and_types()
        for t_name, _ in topic_names_types:
            if t_name not in skip_topics:
                topics.append(t_name)
    return topics



# Copyright 2021 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

from rclpy.qos import qos_check_compatible
from rclpy.qos import QoSCompatibility
from ros2cli.node.strategy import NodeStrategy
from ros2doctor.api import DoctorCheck
from ros2doctor.api import DoctorReport
from ros2doctor.api import get_topic_names
from ros2doctor.api import Report
from ros2doctor.api import Result
from ros2doctor.api.format import doctor_error
from ros2doctor.api.format import doctor_warn


class QoSCompatibilityCheck(DoctorCheck):
    """Check for incompatible QoS profiles in each pub/sub pair."""

    def category(self):
        return 'middleware'

    def check(self):
        """Check publisher and subscriber counts."""
        result = Result()
        to_be_checked = get_topic_names()
        with NodeStrategy(None) as node:
            for topic in to_be_checked:
                for pub in node.get_publishers_info_by_topic(topic):
                    for sub in node.get_subscriptions_info_by_topic(topic):
                        compatibility, reason = qos_check_compatible(
                            pub.qos_profile, sub.qos_profile)
                        reason_message = self._strip_leading_warning_or_error_from_string(reason)
                        if compatibility == QoSCompatibility.WARNING:
                            doctor_warn(f"QoS compatibility warning found on topic '{topic}': "
                                        f'{reason_message}')
                            result.add_warning()
                        elif compatibility == QoSCompatibility.ERROR:
                            doctor_error(f"QoS compatibility error found on topic '{topic}': "
                                         f'{reason_message}')
                            result.add_error()
        return result

    @staticmethod
    def _strip_leading_warning_or_error_from_string(string: str) -> str:
        """
        Remove "warning: " or "error: " (case insensitive) from the beginning of a string.

        If "warning: " or "error: " is not found, the original string is returned.
        """
        re_result = re.search(r'^(?i:warning|error): (.*)', string)
        if re_result:
            assert len(re_result.groups()) == 1
            return re_result.groups()[0]
        else:
            return string


class QoSCompatibilityReport(DoctorReport):
    """Report QoS compatibility related information."""

    def category(self):
        return 'middleware'

    def report(self):
        report = Report('QOS COMPATIBILITY LIST')
        to_be_reported = get_topic_names()
        with NodeStrategy(None) as node:
            for topic in to_be_reported:
                for pub in node.get_publishers_info_by_topic(topic):
                    for sub in node.get_subscriptions_info_by_topic(topic):
                        compatibility, reason = qos_check_compatible(
                            pub.qos_profile, sub.qos_profile)
                        report.add_to_report('topic [type]', f'{topic} [{pub.topic_type}]')
                        report.add_to_report('publisher node', pub.node_name)
                        report.add_to_report('subscriber node', sub.node_name)
                        if compatibility == QoSCompatibility.OK:
                            compatibility_msg = 'OK'
                        else:
                            compatibility_msg = reason
                        report.add_to_report('compatibility status', compatibility_msg)
        if self._is_report_empty(report):
            report.add_to_report('compatibility status', 'No publisher/subscriber pairs found')
        return report

    @staticmethod
    def _is_report_empty(report: Report) -> bool:
        return len(report.items) == 0



# Copyright 2019 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import socket
import struct
import sys

import psutil

from ros2doctor.api import DoctorCheck
from ros2doctor.api import DoctorReport
from ros2doctor.api import Report
from ros2doctor.api import Result
from ros2doctor.api.format import doctor_error
from ros2doctor.api.format import doctor_warn


class InterfaceFlags:
    POSIX_NET_FLAGS = (
        ('UP', 0),
        ('BROADCAST', 1),
        ('DEBUG', 2),
        ('LOOPBACK', 3),

        ('PTP', 4),
        ('NOTRAILERS', 5),
        ('RUNNING', 6),
        ('NOARP', 7),

        ('PROMISC', 8),
        ('ALLMULTI', 9),
        ('MASTER', 10),
        ('SLAVE', 11),

        ('MULTICAST', 12),
        ('PORTSEL', 13),
        ('AUTOMEDIA', 14),
        ('DYNAMIC', 15),
    )

    def __init__(self, interface_name):
        self.flags = -1
        self.flag_list = set()
        self.has_loopback = False
        self.has_non_loopback = False
        self.has_multicast = False
        self.get(interface_name)

    def get(self, interface_name):
        if os.name != 'posix':
            return

        import fcntl

        if sys.platform == 'darwin':
            SIOCGIFFLAGS = 0xc0206911
        else:
            SIOCGIFFLAGS = 0x8913

        # We need to pass a 'struct ifreq' to the SIOCGIFFLAGS ioctl.  Nominally the
        # structure looks like:
        #
        #            struct ifreq {
        #               char ifr_name[IFNAMSIZ]; /* Interface name */
        #               union {
        #                   struct sockaddr ifr_addr;
        #                   struct sockaddr ifr_dstaddr;
        #                   struct sockaddr ifr_broadaddr;
        #                   struct sockaddr ifr_netmask;
        #                   struct sockaddr ifr_hwaddr;
        #                   short           ifr_flags;
        #                   int             ifr_ifindex;
        #                   int             ifr_metric;
        #                   int             ifr_mtu;
        #                   struct ifmap    ifr_map;
        #                   char            ifr_slave[IFNAMSIZ];
        #                   char            ifr_newname[IFNAMSIZ];
        #                   char           *ifr_data;
        #               };
        #           };
        #
        # Where IFNAMSIZ is 16 bytes long.  The caller (that's us) sets the ifr_name
        # to the interface in question, and the call fills in the union.  In the case
        # of this particular ioctl, only the 'ifr_flags' is valid after the call.
        # Either way, we need to provide enough room in the provided structure for the
        # ioctl to fill it out, so we just provide an additional 256 bytes which should
        # be more than enough.

        with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as s:
            try:
                result = fcntl.ioctl(s.fileno(), SIOCGIFFLAGS, interface_name + '\0' * 256)
            except OSError:
                return

        # On success, result contains the structure filled in with the info we asked
        # for (the flags).  We just need to extract it from the correct place.

        self.flags = struct.unpack('H', result[16:18])[0]
        for flagname, bit in self.POSIX_NET_FLAGS:
            if self.flags & (1 << bit):
                self.flag_list.add(flagname)

        if 'LOOPBACK' in self.flag_list:
            self.has_loopback = True
        else:
            self.has_non_loopback = True

        if 'MULTICAST' in self.flag_list:
            self.has_multicast = True

    def __str__(self):
        if self.flags == -1:
            return ''

        output_flags = ','.join(self.flag_list)

        return f'{self.flags}<{output_flags}>'


class NetworkCheck(DoctorCheck):
    """Check network interface configuration for loopback and multicast."""

    def category(self):
        return 'network'

    def check(self):
        """Check network configuration."""
        result = Result()

        has_loopback = False
        has_non_loopback = False
        has_multicast = False
        for interface in psutil.net_if_addrs().keys():
            flags = InterfaceFlags(interface)
            has_loopback |= flags.has_loopback
            has_non_loopback |= flags.has_non_loopback
            has_multicast |= flags.has_multicast

        if os.name != 'posix':
            if not has_loopback and not has_non_loopback:
                # no flags found, otherwise one of them should be True.
                doctor_warn(
                    'Interface flags are not available on Windows. '
                    'Run `ipconfig` to see what interfaces are available.')
                result.add_warning()
                return result
        if not has_loopback:
            doctor_error('No loopback IP address is found.')
            result.add_error()
        if not has_non_loopback:
            doctor_warn('Only loopback IP address is found.')
            result.add_warning()
        if not has_multicast:
            doctor_warn('No multicast IP address is found.')
            result.add_warning()
        return result


class NetworkReport(DoctorReport):
    """Report network configuration."""

    def category(self):
        return 'network'

    def report(self):
        """Print system and ROS network information."""
        if_stats = psutil.net_if_stats()

        network_report = Report('NETWORK CONFIGURATION')

        for interface, addrs in psutil.net_if_addrs().items():
            if_info = {
                'inet': None,
                'inet4': [],
                'ether': None,
                'inet6': [],
                'netmask': None,
                'device': interface,
                'flags': None,
                'mtu': None,
                'broadcast': None,
            }
            for addr in addrs:
                if addr.family == socket.AddressFamily.AF_INET:
                    if_info['inet'] = addr.address
                    if_info['inet4'].append(addr.address)
                    if_info['netmask'] = addr.netmask
                    if_info['broadcast'] = addr.broadcast
                elif addr.family == socket.AddressFamily.AF_INET6:
                    if_info['inet6'].append(addr.address)
                elif 'AF_PACKET' in socket.AddressFamily.__members__:
                    if addr.family == socket.AddressFamily.AF_PACKET:
                        # Loopback reports an all-zero MAC address, which is
                        # bogus and should not be reported
                        if addr.address != '00:00:00:00:00:00':
                            if_info['ether'] = addr.address

            if interface in if_stats:
                if_info['mtu'] = if_stats[interface].mtu

            flags = InterfaceFlags(interface)
            if_info['flags'] = str(flags)

            for k, v in if_info.items():
                if v:
                    network_report.add_to_report(k, v)
        return network_report



# Copyright 2019 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import platform
from typing import Tuple

from ros2doctor.api import DoctorCheck
from ros2doctor.api import DoctorReport
from ros2doctor.api import Report
from ros2doctor.api import Result
from ros2doctor.api.format import doctor_error
from ros2doctor.api.format import doctor_warn

import rosdistro


def _check_platform_helper() -> Tuple[str, dict, dict]:
    """
    Check ROS_DISTRO environment variables and distribution installed.

    :return: string of distro name, dict of distribution info, dict of release platforms info
    """
    distro_name = os.environ.get('ROS_DISTRO')
    if not distro_name:
        doctor_error('ROS_DISTRO is not set.')
        return
    distro_name = distro_name.lower()
    u = rosdistro.get_index_url()
    if not u:
        doctor_error(
            'Unable to access ROSDISTRO_INDEX_URL or DEFAULT_INDEX_URL. '
            'Check network setting to make sure machine is connected to internet.')
        return
    i = rosdistro.get_index(u)
    distro_info = i.distributions.get(distro_name)
    if not distro_info:
        doctor_warn(f'Distribution name {distro_name} is not found')
        return
    try:
        distro_data = rosdistro.get_distribution(i, distro_name).get_data()
    except AttributeError:
        distro_data = ''
    return distro_name, distro_info, distro_data


class PlatformCheck(DoctorCheck):
    """Check system platform against ROSDistro."""

    def category(self):
        return 'platform'

    def check(self):
        """Check system platform against ROS 2 Distro."""
        result = Result()
        distros = _check_platform_helper()
        if not distros:
            doctor_error('Missing rosdistro info. Unable to check platform.')
            result.add_error()
            return result
        distro_name, distro_info, _ = distros

        # check distro status
        if distro_info.get('distribution_status') == 'prerelease':
            doctor_warn(
                f'Distribution {distro_name} is not fully supported or tested. '
                'To get more consistent features, download a stable version at '
                'https://index.ros.org/doc/ros2/Installation/')
            result.add_warning()
        elif distro_info.get('distribution_status') == 'end-of-life':
            doctor_warn(
                f'Distribution {distro_name} is no longer supported or deprecated. '
                'To get the latest features, download the new versions at '
                'https://index.ros.org/doc/ros2/Installation/')
            result.add_warning()
        return result


class PlatformReport(DoctorReport):
    """Output platform report."""

    def category(self):
        return 'platform'

    def report(self):
        platform_name = platform.system()

        # platform info
        platform_report = Report('PLATFORM INFORMATION')
        platform_report.add_to_report('system', platform_name)
        platform_report.add_to_report('platform info', platform.platform())
        if platform_name == 'Darwin':
            platform_report.add_to_report('mac OS version', platform.mac_ver())
        platform_report.add_to_report('release', platform.release())
        platform_report.add_to_report('processor', platform.processor())
        return platform_report


class RosdistroReport(DoctorReport):
    """Output ROSDistro report."""

    def category(self):
        return 'platform'

    def report(self):
        ros_report = Report('ROS 2 INFORMATION')
        distros = _check_platform_helper()
        if not distros:
            return ros_report
        distro_name, distro_info, distro_data = distros
        ros_report.add_to_report('distribution name', distro_name)
        ros_report.add_to_report('distribution type', distro_info.get('distribution_type'))
        ros_report.add_to_report('distribution status', distro_info.get('distribution_status'))
        ros_report.add_to_report('release platforms', distro_data.get('release_platforms'))
        return ros_report
